@@ -278,7 +278,6 @@ Efficient use of soft deletes:
 
 """
 
-import functools
 import itertools
 import logging
 import re
@@ -286,45 +285,21 @@ import time
 
 from oslo.utils import timeutils
 import six
-from sqlalchemy.interfaces import PoolListener
 import sqlalchemy.orm
-from sqlalchemy.pool import NullPool, StaticPool
+from sqlalchemy import pool
 from sqlalchemy.sql.expression import literal_column
 from sqlalchemy.sql.expression import select
 
+from oslo.db._i18n import _LW
 from oslo.db import exception
-from oslo.db.openstack.common.gettextutils import _LW
 from oslo.db import options
+from oslo.db.sqlalchemy import compat
 from oslo.db.sqlalchemy import exc_filters
+from oslo.db.sqlalchemy import utils
 
 LOG = logging.getLogger(__name__)
 
 
-class SqliteForeignKeysListener(PoolListener):
-    """Ensures that the foreign key constraints are enforced in SQLite.
-
-    The foreign key constraints are disabled by default in SQLite,
-    so the foreign key constraints will be enabled here for every
-    database connection
-    """
-    def connect(self, dbapi_con, con_record):
-        dbapi_con.execute('pragma foreign_keys=ON')
-
-
-def _synchronous_switch_listener(dbapi_conn, connection_rec):
-    """Switch sqlite connections to non-synchronous mode."""
-    dbapi_conn.execute("PRAGMA synchronous = OFF")
-
-
-def _add_regexp_listener(dbapi_con, con_record):
-    """Add REGEXP function to sqlite connections."""
-
-    def regexp(expr, item):
-        reg = re.compile(expr)
-        return reg.search(six.text_type(item)) is not None
-    dbapi_con.create_function('regexp', 2, regexp)
-
-
 def _thread_yield(dbapi_con, con_record):
     """Ensure other greenthreads get a chance to be executed.
 
@@ -337,12 +312,22 @@ def _thread_yield(dbapi_con, con_record):
     time.sleep(0)
 
 
-def _begin_ping_listener(connection):
-    """Ping the server at transaction begin.
+def _connect_ping_listener(connection, branch):
+    """Ping the server at connection startup.
 
     Ping the server at transaction begin and transparently reconnect
     if a disconnect exception occurs.
     """
+    if branch:
+        return
+
+    # turn off "close with result".  This can also be accomplished
+    # by branching the connection, however just setting the flag is
+    # more performant and also doesn't get involved with some
+    # connection-invalidation awkardness that occurs (see
+    # https://bitbucket.org/zzzeek/sqlalchemy/issue/3215/)
+    save_should_close_with_result = connection.should_close_with_result
+    connection.should_close_with_result = False
     try:
         # run a SELECT 1.   use a core select() so that
         # any details like that needed by Oracle, DB2 etc. are handled.
@@ -355,65 +340,27 @@ def _begin_ping_listener(connection):
         # new connections assuming they are good now.
         # run the select again to re-validate the Connection.
         connection.scalar(select([1]))
+    finally:
+        connection.should_close_with_result = save_should_close_with_result
 
 
-def _set_session_sql_mode(dbapi_con, connection_rec, sql_mode=None):
-    """Set the sql_mode session variable.
+def _setup_logging(connection_debug=0):
+    """setup_logging function maps SQL debug level to Python log level.
 
-    MySQL supports several server modes. The default is None, but sessions
-    may choose to enable server modes like TRADITIONAL, ANSI,
-    several STRICT_* modes and others.
-
-    Note: passing in '' (empty string) for sql_mode clears
-    the SQL mode for the session, overriding a potentially set
-    server default.
+    Connection_debug is a verbosity of SQL debugging information.
+    0=None(default value),
+    1=Processed only messages with WARNING level or higher
+    50=Processed only messages with INFO level or higher
+    100=Processed only messages with DEBUG level
     """
-
-    cursor = dbapi_con.cursor()
-    cursor.execute("SET SESSION sql_mode = %s", [sql_mode])
-
-
-def _mysql_get_effective_sql_mode(engine):
-    """Returns the effective SQL mode for connections from the engine pool.
-
-    Returns ``None`` if the mode isn't available, otherwise returns the mode.
-
-    """
-    # Get the real effective SQL mode. Even when unset by
-    # our own config, the server may still be operating in a specific
-    # SQL mode as set by the server configuration.
-    # Also note that the checkout listener will be called on execute to
-    # set the mode if it's registered.
-    row = engine.execute("SHOW VARIABLES LIKE 'sql_mode'").fetchone()
-    if row is None:
-        return
-    return row[1]
-
-
-def _mysql_check_effective_sql_mode(engine):
-    """Logs a message based on the effective SQL mode for MySQL connections."""
-    realmode = _mysql_get_effective_sql_mode(engine)
-
-    if realmode is None:
-        LOG.warning(_LW('Unable to detect effective SQL mode'))
-        return
-
-    LOG.debug('MySQL server mode set to %s', realmode)
-    # 'TRADITIONAL' mode enables several other modes, so
-    # we need a substring match here
-    if not ('TRADITIONAL' in realmode.upper() or
-            'STRICT_ALL_TABLES' in realmode.upper()):
-        LOG.warning(_LW("MySQL SQL mode is '%s', "
-                        "consider enabling TRADITIONAL or STRICT_ALL_TABLES"),
-                    realmode)
-
-
-def _mysql_set_mode_callback(engine, sql_mode):
-    if sql_mode is not None:
-        mode_callback = functools.partial(_set_session_sql_mode,
-                                          sql_mode=sql_mode)
-        sqlalchemy.event.listen(engine, 'connect', mode_callback)
-    _mysql_check_effective_sql_mode(engine)
+    if connection_debug >= 0:
+        logger = logging.getLogger('sqlalchemy.engine')
+        if connection_debug >= 100:
+            logger.setLevel(logging.DEBUG)
+        elif connection_debug >= 50:
+            logger.setLevel(logging.INFO)
+        else:
+            logger.setLevel(logging.WARNING)
 
 
 def create_engine(sql_connection, sqlite_fk=False, mysql_sql_mode=None,
@@ -424,71 +371,163 @@ def create_engine(sql_connection, sqlite_fk=False, mysql_sql_mode=None,
                   thread_checkin=True):
     """Return a new SQLAlchemy engine."""
 
-    connection_dict = sqlalchemy.engine.url.make_url(sql_connection)
+    url = sqlalchemy.engine.url.make_url(sql_connection)
 
     engine_args = {
         "pool_recycle": idle_timeout,
         'convert_unicode': True,
     }
 
-    if connection_debug >= 0:
-        # Map SQL debug level to Python log level
-        logger = logging.getLogger('sqlalchemy.engine')
-        if connection_debug >= 100:
-            logger.setLevel(logging.DEBUG)
-        elif connection_debug >= 50:
-            logger.setLevel(logging.INFO)
-        else:
-            logger.setLevel(logging.WARNING)
+    _setup_logging(connection_debug)
 
-    if "sqlite" in connection_dict.drivername:
-        if sqlite_fk:
-            engine_args["listeners"] = [SqliteForeignKeysListener()]
-        engine_args["poolclass"] = NullPool
+    _init_connection_args(
+        url, engine_args,
+        sqlite_fk=sqlite_fk,
+        max_pool_size=max_pool_size,
+        max_overflow=max_overflow,
+        pool_timeout=pool_timeout
+    )
 
-        if sql_connection == "sqlite://":
-            engine_args["poolclass"] = StaticPool
-            engine_args["connect_args"] = {'check_same_thread': False}
-    else:
+    engine = sqlalchemy.create_engine(url, **engine_args)
+
+    _init_events(
+        engine,
+        mysql_sql_mode=mysql_sql_mode,
+        sqlite_synchronous=sqlite_synchronous,
+        sqlite_fk=sqlite_fk,
+        thread_checkin=thread_checkin,
+        connection_trace=connection_trace
+    )
+
+    # register alternate exception handler
+    exc_filters.register_engine(engine)
+
+    # register engine connect handler
+    compat.engine_connect(engine, _connect_ping_listener)
+
+    # initial connect + test
+    _test_connection(engine, max_retries, retry_interval)
+
+    return engine
+
+
+@utils.dispatch_for_dialect('*', multiple=True)
+def _init_connection_args(
+    url, engine_args,
+    max_pool_size=None, max_overflow=None, pool_timeout=None, **kw):
+
+    pool_class = url.get_dialect().get_pool_class(url)
+    if issubclass(pool_class, pool.QueuePool):
         if max_pool_size is not None:
             engine_args['pool_size'] = max_pool_size
         if max_overflow is not None:
             engine_args['max_overflow'] = max_overflow
         if pool_timeout is not None:
             engine_args['pool_timeout'] = pool_timeout
-        if connection_dict.get_dialect().driver == 'mysqlconnector':
-            # mysqlconnector engine (<1.0) incorrectly defaults to
-            # raise_on_warnings=True
-            #  https://bitbucket.org/zzzeek/sqlalchemy/issue/2515
-            engine_args['connect_args'] = {'raise_on_warnings': False}
 
-    engine = sqlalchemy.create_engine(sql_connection, **engine_args)
 
-    if thread_checkin:
-        sqlalchemy.event.listen(engine, 'checkin', _thread_yield)
+@_init_connection_args.dispatch_for("sqlite")
+def _init_connection_args(url, engine_args, **kw):
+    pool_class = url.get_dialect().get_pool_class(url)
+    # singletonthreadpool is used for :memory: connections;
+    # replace it with StaticPool.
+    if issubclass(pool_class, pool.SingletonThreadPool):
+        engine_args["poolclass"] = pool.StaticPool
+        engine_args["connect_args"] = {'check_same_thread': False}
 
-    if engine.name == 'mysql':
-        if mysql_sql_mode is not None:
-            _mysql_set_mode_callback(engine, mysql_sql_mode)
-    elif 'sqlite' in connection_dict.drivername:
-        if not sqlite_synchronous:
-            sqlalchemy.event.listen(engine, 'connect',
-                                    _synchronous_switch_listener)
-        sqlalchemy.event.listen(engine, 'connect', _add_regexp_listener)
+
+@_init_connection_args.dispatch_for("mysql+mysqlconnector")
+def _init_connection_args(url, engine_args, **kw):
+    # mysqlconnector engine (<1.0) incorrectly defaults to
+    # raise_on_warnings=True
+    #  https://bitbucket.org/zzzeek/sqlalchemy/issue/2515
+    if "raise_on_warnings" not in url.query:
+        engine_args['connect_args'] = {'raise_on_warnings': False}
+
+
+@utils.dispatch_for_dialect('*', multiple=True)
+def _init_events(engine, thread_checkin=True, connection_trace=False, **kw):
+    """Set up event listeners for all database backends."""
 
     if connection_trace:
         _add_trace_comments(engine)
 
-    # register alternate exception handler
-    exc_filters.register_engine(engine)
+    if thread_checkin:
+        sqlalchemy.event.listen(engine, 'checkin', _thread_yield)
 
-    # register on begin handler
-    sqlalchemy.event.listen(engine, "begin", _begin_ping_listener)
 
-    # initial connect + test
-    _test_connection(engine, max_retries, retry_interval)
+@_init_events.dispatch_for("mysql")
+def _init_events(engine, mysql_sql_mode=None, **kw):
+    """Set up event listeners for MySQL."""
 
-    return engine
+    if mysql_sql_mode is not None:
+        @sqlalchemy.event.listens_for(engine, "connect")
+        def _set_session_sql_mode(dbapi_con, connection_rec):
+            cursor = dbapi_con.cursor()
+            cursor.execute("SET SESSION sql_mode = %s", [mysql_sql_mode])
+
+    realmode = engine.execute("SHOW VARIABLES LIKE 'sql_mode'").fetchone()
+    if realmode is None:
+        LOG.warning(_LW('Unable to detect effective SQL mode'))
+    else:
+        realmode = realmode[1]
+        LOG.debug('MySQL server mode set to %s', realmode)
+        if 'TRADITIONAL' not in realmode.upper() and \
+            'STRICT_ALL_TABLES' not in realmode.upper():
+            LOG.warning(
+                _LW(
+                    "MySQL SQL mode is '%s', "
+                    "consider enabling TRADITIONAL or STRICT_ALL_TABLES"),
+                realmode)
+
+
+@_init_events.dispatch_for("sqlite")
+def _init_events(engine, sqlite_synchronous=True, sqlite_fk=False, **kw):
+    """Set up event listeners for SQLite.
+
+    This includes several settings made on connections as they are
+    created, as well as transactional control extensions.
+
+    """
+
+    def regexp(expr, item):
+        reg = re.compile(expr)
+        return reg.search(six.text_type(item)) is not None
+
+    @sqlalchemy.event.listens_for(engine, "connect")
+    def _sqlite_connect_events(dbapi_con, con_record):
+
+        # Add REGEXP functionality on SQLite connections
+        dbapi_con.create_function('regexp', 2, regexp)
+
+        if not sqlite_synchronous:
+            # Switch sqlite connections to non-synchronous mode
+            dbapi_con.execute("PRAGMA synchronous = OFF")
+
+        # Disable pysqlite's emitting of the BEGIN statement entirely.
+        # Also stops it from emitting COMMIT before any DDL.
+        # below, we emit BEGIN ourselves.
+        # see http://docs.sqlalchemy.org/en/rel_0_9/dialects/\
+        # sqlite.html#serializable-isolation-savepoints-transactional-ddl
+        dbapi_con.isolation_level = None
+
+        if sqlite_fk:
+            # Ensures that the foreign key constraints are enforced in SQLite.
+            dbapi_con.execute('pragma foreign_keys=ON')
+
+    @sqlalchemy.event.listens_for(engine, "begin")
+    def _sqlite_emit_begin(conn):
+        # emit our own BEGIN, checking for existing
+        # transactional state
+        if 'in_transaction' not in conn.info:
+            conn.execute("BEGIN")
+            conn.info['in_transaction'] = True
+
+    @sqlalchemy.event.listens_for(engine, "rollback")
+    @sqlalchemy.event.listens_for(engine, "commit")
+    def _sqlite_end_transaction(conn):
+        # remove transactional marker
+        conn.info.pop('in_transaction', None)
 
 
 def _test_connection(engine, max_retries, retry_interval):
