@@ -30,8 +30,9 @@ from sqlalchemy import schema
 import sqlalchemy.sql.expression as expr
 import sqlalchemy.types as types
 
+from oslo.db._i18n import _LE
 from oslo.db import exception as exc
-from oslo.db.openstack.common.gettextutils import _LE
+from oslo.db.sqlalchemy import utils
 
 LOG = logging.getLogger(__name__)
 
@@ -44,12 +45,11 @@ class WalkVersionsMixin(object):
     abstract class mixin. `INIT_VERSION`, `REPOSITORY` and `migration_api`
     attributes must be implemented in subclasses.
 
-    .. _auxiliary-dynamic-methods:
+    .. _auxiliary-dynamic-methods: Auxiliary Methods
 
-    Auxiliary methods
-    -----------------
+    Auxiliary Methods:
 
-    `_migrate_up` and `_migrate_down` instance methods of the class can be
+    `migrate_up` and `migrate_down` instance methods of the class can be
     used with auxiliary methods named `_pre_upgrade_<revision_id>`,
     `_check_<revision_id>`, `_post_downgrade_<revision_id>`. The methods
     intended to check applied changes for correctness of data operations.
@@ -58,16 +58,16 @@ class WalkVersionsMixin(object):
     `_pre_upgrade_<revision_id>`, `_check_<revision_id>`,
     `_post_downgrade_<revision_id>` implementation:
 
-        * `_pre_upgrade_<revision_id>`: provide a data appropriate to a
-        next revision. Should be used an id of revision which going to be
-        applied.
+        * `_pre_upgrade_<revision_id>`: provide a data appropriate to
+           a next revision. Should be used an id of revision which
+           going to be applied.
 
-        * `_check_<revision_id>`: Insert, select, delete operations with
-        newly applied changes. The data provided by
-        `_pre_upgrade_<revision_id>` will be used.
+        * `_check_<revision_id>`: Insert, select, delete operations
+           with newly applied changes. The data provided by
+           `_pre_upgrade_<revision_id>` will be used.
 
-        *`_post_downgrade_<revision_id>`: check for absence (inability to
-        use) changes provided by reverted revision.
+        * `_post_downgrade_<revision_id>`: check for absence
+          (inability to use) changes provided by reverted revision.
 
     Execution order of auxiliary methods when revision is upgrading:
 
@@ -78,6 +78,7 @@ class WalkVersionsMixin(object):
         `downgrade` => `_post_downgrade_###`
 
     .. _migrate: https://sqlalchemy-migrate.readthedocs.org/en/latest/
+
     """
 
     @abc.abstractproperty
@@ -126,14 +127,38 @@ class WalkVersionsMixin(object):
     def _walk_versions(self, snake_walk=False, downgrade=True):
         """Check if migration upgrades and downgrades successfully.
 
+        DEPRECATED: this function is deprecated and will be removed from
+        oslo.db in a few releases. Please use walk_versions() method instead.
+        """
+        self.walk_versions(snake_walk, downgrade)
+
+    def _migrate_down(self, version, with_data=False):
+        """Migrate down to a previous version of the db.
+
+        DEPRECATED: this function is deprecated and will be removed from
+        oslo.db in a few releases. Please use migrate_down() method instead.
+        """
+        return self.migrate_down(version, with_data)
+
+    def _migrate_up(self, version, with_data=False):
+        """Migrate up to a new version of the db.
+
+        DEPRECATED: this function is deprecated and will be removed from
+        oslo.db in a few releases. Please use migrate_up() method instead.
+        """
+        self.migrate_up(version, with_data)
+
+    def walk_versions(self, snake_walk=False, downgrade=True):
+        """Check if migration upgrades and downgrades successfully.
+
         Determine the latest version script from the repo, then
         upgrade from 1 through to the latest, with no data
         in the databases. This just checks that the schema itself
         upgrades successfully.
 
-        `_walk_versions` calls `_migrate_up` and `_migrate_down` with
+        `walk_versions` calls `migrate_up` and `migrate_down` with
         `with_data` argument to check changes with data, but these methods
-        can be called without any extra check outside of `_walk_versions`
+        can be called without any extra check outside of `walk_versions`
         method.
 
         :param snake_walk: enables checking that each individual migration can
@@ -164,31 +189,31 @@ class WalkVersionsMixin(object):
 
         for version in versions:
             # upgrade -> downgrade -> upgrade
-            self._migrate_up(version, with_data=True)
+            self.migrate_up(version, with_data=True)
             if snake_walk:
-                downgraded = self._migrate_down(version - 1, with_data=True)
+                downgraded = self.migrate_down(version - 1, with_data=True)
                 if downgraded:
-                    self._migrate_up(version)
+                    self.migrate_up(version)
 
         if downgrade:
             # Now walk it back down to 0 from the latest, testing
             # the downgrade paths.
             for version in reversed(versions):
                 # downgrade -> upgrade -> downgrade
-                downgraded = self._migrate_down(version - 1)
+                downgraded = self.migrate_down(version - 1)
 
                 if snake_walk and downgraded:
-                    self._migrate_up(version)
-                    self._migrate_down(version - 1)
+                    self.migrate_up(version)
+                    self.migrate_down(version - 1)
 
-    def _migrate_down(self, version, with_data=False):
+    def migrate_down(self, version, with_data=False):
         """Migrate down to a previous version of the db.
 
         :param version: id of revision to downgrade.
         :type version: str
         :keyword with_data: Whether to verify the absence of changes from
             migration(s) being downgraded, see
-            :ref:`auxiliary-dynamic-methods`.
+            :ref:`auxiliary-dynamic-methods <Auxiliary Methods>`.
         :type with_data: Bool
         """
 
@@ -214,13 +239,13 @@ class WalkVersionsMixin(object):
 
         return True
 
-    def _migrate_up(self, version, with_data=False):
+    def migrate_up(self, version, with_data=False):
         """Migrate up to a new version of the db.
 
         :param version: id of revision to upgrade.
         :type version: str
         :keyword with_data: Whether to verify the applied changes with data,
-            see :ref:`auxiliary-dynamic-methods`.
+            see :ref:`auxiliary-dynamic-methods <Auxiliary Methods>`.
         :type with_data: Bool
         """
         # NOTE(sdague): try block is here because it's impossible to debug
@@ -346,19 +371,34 @@ class ModelsMigrationsSync(object):
         :param rendered_meta_def: rendered column default value (from model)
 
         """
+        return self._compare_server_default(ctxt.bind, meta_col, insp_def,
+                                            meta_def)
 
-        if (ctxt.dialect.name == 'mysql' and
-                issubclass(type(meta_col.type), sqlalchemy.Boolean)):
+    @utils.DialectFunctionDispatcher.dispatch_for_dialect("*")
+    def _compare_server_default(bind, meta_col, insp_def, meta_def):
+        pass
 
+    @_compare_server_default.dispatch_for('mysql')
+    def _compare_server_default(bind, meta_col, insp_def, meta_def):
+        if isinstance(meta_col.type, sqlalchemy.Boolean):
             if meta_def is None or insp_def is None:
                 return meta_def != insp_def
-
             return not (
                 isinstance(meta_def.arg, expr.True_) and insp_def == "'1'" or
                 isinstance(meta_def.arg, expr.False_) and insp_def == "'0'"
             )
 
-        return None  # tells alembic to use the default comparison method
+        if isinstance(meta_col.type, sqlalchemy.Integer):
+            if meta_def is None or insp_def is None:
+                return meta_def != insp_def
+            return meta_def.arg != insp_def.split("'")[1]
+
+    @_compare_server_default.dispatch_for('postgresql')
+    def _compare_server_default(bind, meta_col, insp_def, meta_def):
+        if isinstance(meta_col.type, sqlalchemy.String):
+            if meta_def is None or insp_def is None:
+                return meta_def != insp_def
+            return insp_def != "'%s'::character varying" % meta_def.arg
 
     def _cleanup(self):
         engine = self.get_engine()
