@@ -16,6 +16,7 @@
 #    under the License.
 
 """Unit tests for SQLAlchemy specific code."""
+
 import logging
 
 import fixtures
@@ -24,6 +25,7 @@ from oslo.config import cfg
 from oslotest import base as oslo_test
 import sqlalchemy
 from sqlalchemy import Column, MetaData, Table
+from sqlalchemy.engine import url
 from sqlalchemy import Integer, String
 from sqlalchemy.ext.declarative import declarative_base
 
@@ -84,6 +86,87 @@ class RegexpFilterTestCase(test_base.DbTestCase):
         self._test_regexp_filter(u'♦', [])
 
 
+class SQLiteSavepointTest(test_base.DbTestCase):
+    def setUp(self):
+        super(SQLiteSavepointTest, self).setUp()
+        meta = MetaData()
+        self.test_table = Table(
+            "test_table", meta,
+            Column('id', Integer, primary_key=True),
+            Column('data', String(10)))
+        self.test_table.create(self.engine)
+        self.addCleanup(self.test_table.drop, self.engine)
+
+    def test_plain_transaction(self):
+        conn = self.engine.connect()
+        trans = conn.begin()
+        conn.execute(
+            self.test_table.insert(),
+            {'data': 'data 1'}
+        )
+        self.assertEqual(
+            [(1, 'data 1')],
+            self.engine.execute(
+                self.test_table.select().
+                order_by(self.test_table.c.id)
+            ).fetchall()
+        )
+        trans.rollback()
+        self.assertEqual(
+            0,
+            self.engine.scalar(self.test_table.count())
+        )
+
+    def test_savepoint_middle(self):
+        with self.engine.begin() as conn:
+            conn.execute(
+                self.test_table.insert(),
+                {'data': 'data 1'}
+            )
+
+            savepoint = conn.begin_nested()
+            conn.execute(
+                self.test_table.insert(),
+                {'data': 'data 2'}
+            )
+            savepoint.rollback()
+
+            conn.execute(
+                self.test_table.insert(),
+                {'data': 'data 3'}
+            )
+
+        self.assertEqual(
+            [(1, 'data 1'), (2, 'data 3')],
+            self.engine.execute(
+                self.test_table.select().
+                order_by(self.test_table.c.id)
+            ).fetchall()
+        )
+
+    def test_savepoint_beginning(self):
+        with self.engine.begin() as conn:
+            savepoint = conn.begin_nested()
+            conn.execute(
+                self.test_table.insert(),
+                {'data': 'data 1'}
+            )
+            savepoint.rollback()
+
+            conn.execute(
+                self.test_table.insert(),
+                {'data': 'data 2'}
+            )
+
+        self.assertEqual(
+            [(1, 'data 2')],
+            self.engine.execute(
+                self.test_table.select().
+                order_by(self.test_table.c.id)
+            ).fetchall()
+        )
+
+
 class FakeDBAPIConnection():
     def cursor(self):
         return FakeCursor()
@@ -294,175 +377,248 @@ class EngineFacadeTestCase(oslo_test.BaseTestCase):
         self.assertEqual(master_path, str(slave_session.bind.url))
 
 
-class MysqlSetCallbackTest(oslo_test.BaseTestCase):
+class SQLiteConnectTest(oslo_test.BaseTestCase):
 
-    class FakeCursor(object):
-        def __init__(self, execs):
-            self._execs = execs
+    def _fixture(self, **kw):
+        return session.create_engine("sqlite://", **kw)
 
-        def execute(self, sql, arg):
-            self._execs.append(sql % arg)
+    def test_sqlite_fk_listener(self):
+        engine = self._fixture(sqlite_fk=True)
+        self.assertEqual(
+            engine.scalar("pragma foreign_keys"),
+            1
+        )
 
-    class FakeDbapiCon(object):
-        def __init__(self, execs):
-            self._execs = execs
+        engine = self._fixture(sqlite_fk=False)
 
-        def cursor(self):
-            return MysqlSetCallbackTest.FakeCursor(self._execs)
+        self.assertEqual(
+            engine.scalar("pragma foreign_keys"),
+            0
+        )
 
-    class FakeResultSet(object):
-        def __init__(self, realmode):
-            self._realmode = realmode
+    def test_sqlite_synchronous_listener(self):
+        engine = self._fixture()
 
-        def fetchone(self):
-            return ['ignored', self._realmode]
+        # "The default setting is synchronous=FULL." (e.g. 2)
+        # http://www.sqlite.org/pragma.html#pragma_synchronous
+        self.assertEqual(
+            engine.scalar("pragma synchronous"),
+            2
+        )
 
-    class FakeEngine(object):
-        def __init__(self, realmode=None):
-            self._cbs = {}
-            self._execs = []
-            self._realmode = realmode
-            self._connected = False
+        engine = self._fixture(sqlite_synchronous=False)
 
-        def set_callback(self, name, cb):
-            self._cbs[name] = cb
+        self.assertEqual(
+            engine.scalar("pragma synchronous"),
+            0
+        )
 
-        def connect(self, **kwargs):
-            cb = self._cbs.get('connect', lambda *x, **y: None)
-            dbapi_con = MysqlSetCallbackTest.FakeDbapiCon(self._execs)
-            connection_rec = None  # Not used.
-            cb(dbapi_con, connection_rec)
 
-        def execute(self, sql):
-            if not self._connected:
-                self.connect()
-                self._connected = True
-            self._execs.append(sql)
-            return MysqlSetCallbackTest.FakeResultSet(self._realmode)
+class MysqlConnectTest(test_base.MySQLOpportunisticTestCase):
 
-    def stub_listen(engine, name, cb):
-        engine.set_callback(name, cb)
+    def _fixture(self, sql_mode):
+        return session.create_engine(self.engine.url, mysql_sql_mode=sql_mode)
 
-    @mock.patch.object(sqlalchemy.event, 'listen', side_effect=stub_listen)
-    def _call_set_callback(self, listen_mock, sql_mode=None, realmode=None):
-        engine = self.FakeEngine(realmode=realmode)
-
-        self.stream = self.useFixture(fixtures.FakeLogger(
-            format="%(levelname)8s [%(name)s] %(message)s",
-            level=logging.DEBUG,
-            nuke_handlers=True
-        ))
-
-        session._mysql_set_mode_callback(engine, sql_mode=sql_mode)
-        return engine
+    def _assert_sql_mode(self, engine, sql_mode_present, sql_mode_non_present):
+        mode = engine.execute("SHOW VARIABLES LIKE 'sql_mode'").fetchone()[1]
+        self.assertTrue(
+            sql_mode_present in mode
+        )
+        if sql_mode_non_present:
+            self.assertTrue(
+                sql_mode_non_present not in mode
+            )
 
     def test_set_mode_traditional(self):
-        # If _mysql_set_mode_callback is called with an sql_mode, then the SQL
-        # mode is set on the connection.
-
-        engine = self._call_set_callback(sql_mode='TRADITIONAL')
-
-        exp_calls = [
-            "SET SESSION sql_mode = ['TRADITIONAL']",
-            "SHOW VARIABLES LIKE 'sql_mode'"
-        ]
-        self.assertEqual(exp_calls, engine._execs)
+        engine = self._fixture(sql_mode='TRADITIONAL')
+        self._assert_sql_mode(engine, "TRADITIONAL", "ANSI")
 
     def test_set_mode_ansi(self):
-        # If _mysql_set_mode_callback is called with an sql_mode, then the SQL
-        # mode is set on the connection.
-
-        engine = self._call_set_callback(sql_mode='ANSI')
-
-        exp_calls = [
-            "SET SESSION sql_mode = ['ANSI']",
-            "SHOW VARIABLES LIKE 'sql_mode'"
-        ]
-        self.assertEqual(exp_calls, engine._execs)
+        engine = self._fixture(sql_mode='ANSI')
+        self._assert_sql_mode(engine, "ANSI", "TRADITIONAL")
 
     def test_set_mode_no_mode(self):
         # If _mysql_set_mode_callback is called with sql_mode=None, then
         # the SQL mode is NOT set on the connection.
 
-        engine = self._call_set_callback()
+        expected = self.engine.execute(
+            "SHOW VARIABLES LIKE 'sql_mode'").fetchone()[1]
 
-        exp_calls = [
-            "SHOW VARIABLES LIKE 'sql_mode'"
-        ]
-        self.assertEqual(exp_calls, engine._execs)
+        engine = self._fixture(sql_mode=None)
+        self._assert_sql_mode(engine, expected, None)
 
     def test_fail_detect_mode(self):
         # If "SHOW VARIABLES LIKE 'sql_mode'" results in no row, then
         # we get a log indicating can't detect the mode.
 
-        self._call_set_callback()
+        log = self.useFixture(fixtures.FakeLogger(level=logging.WARN))
+
+        engine = self._fixture(sql_mode=None)
+
+        @sqlalchemy.event.listens_for(
+            engine, "before_cursor_execute", retval=True)
+        def replace_stmt(
+            conn, cursor, statement, parameters,
+            context, executemany):
+            if "SHOW VARIABLES LIKE 'sql_mode'" in statement:
+                statement = "SHOW VARIABLES LIKE 'i_dont_exist'"
+            return statement, parameters
+
+        session._init_events.dispatch_on_drivername("mysql")(engine)
 
         self.assertIn('Unable to detect effective SQL mode',
-                      self.stream.output)
+                      log.output)
 
     def test_logs_real_mode(self):
         # If "SHOW VARIABLES LIKE 'sql_mode'" results in a value, then
         # we get a log with the value.
 
-        self._call_set_callback(realmode='SOMETHING')
+        log = self.useFixture(fixtures.FakeLogger(level=logging.DEBUG))
 
-        self.assertIn('MySQL server mode set to SOMETHING',
-                      self.stream.output)
+        engine = self._fixture(sql_mode='TRADITIONAL')
+
+        actual_mode = engine.execute(
+            "SHOW VARIABLES LIKE 'sql_mode'").fetchone()[1]
+
+        self.assertIn('MySQL server mode set to %s' % actual_mode,
+                      log.output)
 
     def test_warning_when_not_traditional(self):
         # If "SHOW VARIABLES LIKE 'sql_mode'" results in a value that doesn't
         # include 'TRADITIONAL', then a warning is logged.
 
-        self._call_set_callback(realmode='NOT_TRADIT')
+        log = self.useFixture(fixtures.FakeLogger(level=logging.WARN))
+        self._fixture(sql_mode='ANSI')
 
         self.assertIn("consider enabling TRADITIONAL or STRICT_ALL_TABLES",
-                      self.stream.output)
+                      log.output)
 
     def test_no_warning_when_traditional(self):
         # If "SHOW VARIABLES LIKE 'sql_mode'" results in a value that includes
         # 'TRADITIONAL', then no warning is logged.
 
-        self._call_set_callback(realmode='TRADITIONAL')
+        log = self.useFixture(fixtures.FakeLogger(level=logging.WARN))
+        self._fixture(sql_mode='TRADITIONAL')
 
         self.assertNotIn("consider enabling TRADITIONAL or STRICT_ALL_TABLES",
-                         self.stream.output)
+                         log.output)
 
     def test_no_warning_when_strict_all_tables(self):
         # If "SHOW VARIABLES LIKE 'sql_mode'" results in a value that includes
         # 'STRICT_ALL_TABLES', then no warning is logged.
 
-        self._call_set_callback(realmode='STRICT_ALL_TABLES')
+        log = self.useFixture(fixtures.FakeLogger(level=logging.WARN))
+        self._fixture(sql_mode='TRADITIONAL')
 
         self.assertNotIn("consider enabling TRADITIONAL or STRICT_ALL_TABLES",
-                         self.stream.output)
+                         log.output)
 
-    def test_multiple_executes(self):
-        # We should only set the sql_mode on a connection once.
 
-        engine = self._call_set_callback(sql_mode='TRADITIONAL',
-                                         realmode='TRADITIONAL')
+class CreateEngineTest(oslo_test.BaseTestCase):
+    """Test that dialect-specific arguments/ listeners are set up correctly.
 
-        engine.execute('SELECT * FROM foo')
-        engine.execute('SELECT * FROM bar')
+    """
 
-        exp_calls = [
-            "SET SESSION sql_mode = ['TRADITIONAL']",
-            "SHOW VARIABLES LIKE 'sql_mode'",
-            "SELECT * FROM foo",
-            "SELECT * FROM bar",
-        ]
-        self.assertEqual(exp_calls, engine._execs)
+    def test_queuepool_args(self):
+        args = {}
+        session._init_connection_args(
+            url.make_url("mysql://u:p@host/test"), args,
+            max_pool_size=10, max_overflow=10)
+        self.assertEqual(args['pool_size'], 10)
+        self.assertEqual(args['max_overflow'], 10)
+
+    def test_sqlite_memory_pool_args(self):
+        for _url in ("sqlite://", "sqlite:///:memory:"):
+            args = {}
+            session._init_connection_args(
+                url.make_url(_url), args,
+                max_pool_size=10, max_overflow=10)
+
+            # queuepool arguments are not peresnet
+            self.assertTrue(
+                'pool_size' not in args)
+            self.assertTrue(
+                'max_overflow' not in args)
+
+            self.assertEqual(
+                args['connect_args'],
+                {'check_same_thread': False}
+            )
+
+            # due to memory connection
+            self.assertTrue('poolclass' in args)
+
+    def test_sqlite_file_pool_args(self):
+        args = {}
+        session._init_connection_args(
+            url.make_url("sqlite:///somefile.db"), args,
+            max_pool_size=10, max_overflow=10)
+
+        # queuepool arguments are not peresnet
+        self.assertTrue('pool_size' not in args)
+        self.assertTrue(
+            'max_overflow' not in args)
+
+        self.assertTrue(
+            'connect_args' not in args
+        )
+
+        # NullPool is the default for file based connections,
+        # no need to specify this
+        self.assertTrue('poolclass' not in args)
+
+    def test_mysql_connect_args_default(self):
+        args = {}
+        session._init_connection_args(
+            url.make_url("mysql+mysqldb://u:p@host/test"), args)
+        self.assertTrue(
+            'connect_args' not in args
+        )
+
+    def test_mysqlconnector_raise_on_warnings_default(self):
+        args = {}
+        session._init_connection_args(
+            url.make_url("mysql+mysqlconnector://u:p@host/test"),
+            args)
+        self.assertEqual(
+            args,
+            {'connect_args': {'raise_on_warnings': False}}
+        )
+
+    def test_mysqlconnector_raise_on_warnings_override(self):
+        args = {}
+        session._init_connection_args(
+            url.make_url(
+                "mysql+mysqlconnector://u:p@host/test"
+                "?raise_on_warnings=true"),
+            args
+        )
+
+        self.assertTrue(
+            'connect_args' not in args
+        )
+
+    def test_thread_checkin(self):
+        with mock.patch("sqlalchemy.event.listens_for"):
+            with mock.patch("sqlalchemy.event.listen") as listen_evt:
+                session._init_events.dispatch_on_drivername(
+                    "sqlite")(mock.Mock())
+
+        self.assertEqual(
+            listen_evt.mock_calls[0][1][-1],
+            session._thread_yield
+        )
 
 
 class PatchStacktraceTest(test_base.DbTestCase):
 
     def test_trace(self):
         engine = self.engine
-        engine.connect()
+        session._add_trace_comments(engine)
+        conn = engine.connect()
         with mock.patch.object(engine.dialect, "do_execute") as mock_exec:
-            session._add_trace_comments(engine)
 
-            engine.execute("select * from table")
+            conn.execute("select * from table")
 
         call = mock_exec.mock_calls[0]
 
